/*
TX fifo control - Copyright (C) KI6ZUM 2015
Copyright (C) 2015 by Jonathan Naylor G4KLX

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
Boston, MA  02110-1301, USA.
*/

#include "SampleRB.h"

CSampleRB::CSampleRB() :
m_samples(),
m_control(),
m_head(0U),
m_tail(0U),
m_overflow(false)
{
}

uint16_t CSampleRB::getSpace() const
{
  if (m_tail == m_head)
    return SAMPLE_RINGBUFFER_SIZE;
  else if (m_tail < m_head)
    return SAMPLE_RINGBUFFER_SIZE - m_head + m_tail;
  else
    return m_tail - m_head;
}

uint16_t CSampleRB::getData() const
{
  if (m_tail == m_head)
    return 0U;
  else if (m_tail < m_head)
    return m_head - m_tail;
  else
    return SAMPLE_RINGBUFFER_SIZE - m_tail + m_head;
}

void CSampleRB::put(uint16_t sample, uint8_t control)
{
  m_samples[m_head] = sample;
  m_control[m_head] = control;

  m_head = (m_head + 1U) % SAMPLE_RINGBUFFER_SIZE;

  if (m_head == m_tail)
    m_overflow = true;
}

void CSampleRB::get(uint16_t& sample, uint8_t& control)
{
  sample  = m_samples[m_tail];
  control = m_control[m_tail];

  m_tail = (m_tail + 1U) % SAMPLE_RINGBUFFER_SIZE;
}

bool CSampleRB::hasOverflow()
{
  bool overflow = m_overflow;
  
  m_overflow = false;
  
  return overflow;
}

